﻿#pragma once
#include <Utils/Utils.h>
typedef Event<void(class Control*, HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)> EventHandler;
class Control
{
protected:
	static HFONT DefaultFont();
	static LRESULT CALLBACK CAPTURED_MESSAGE_PROCESS(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT(*MESSAGE_PROCESS)(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) = NULL;
public:
	HWND m_Handle = NULL;
	EventHandler OnClicked;
	EventHandler OnDoubleClicked;
	EventHandler OnChanged;
	EventHandler OnEditChanged;
	EventHandler OnSelectionChanged;

	EventHandler OnMouseLeftButtonDown;
	EventHandler OnMouseLeftButtonUp;

	EventHandler OnMouseMiddleButtonDown;
	EventHandler OnMouseMiddleButtonUp;

	EventHandler OnMouseRightButtonDown;
	EventHandler OnMouseRightButtonUp;

	EventHandler OnMouseMove;
	EventHandler OnMouseWheel;

	EventHandler OnKeyDown;
	EventHandler OnKeyUp;
	EventHandler OnGotFocus;
	EventHandler OnLostFocus;

	EventHandler OnSizeChanged;
	EventHandler OnMoved;
	Control();

	PROPERTY(HWND, Parent);
	GET(HWND, Parent);
	SET(HWND, Parent);

	PROPERTY(Control*, ParentControl);
	GET(Control*, ParentControl);
	SET(Control*, ParentControl);

	PROPERTY(std::wstring, Text);
	GET(std::wstring, Text);
	SET(std::wstring, Text);

	PROPERTY(DWORD, Style);
	GET(DWORD, Style);
	SET(DWORD, Style);

	PROPERTY(DWORD, StyleEx);
	GET(DWORD, StyleEx);
	SET(DWORD, StyleEx);

	PROPERTY(std::wstring, Class);
	GET(std::wstring, Class);

	PROPERTY(int, Width);
	GET(int, Width);
	SET(int, Width);
	PROPERTY(int, Height);
	GET(int, Height);
	SET(int, Height);

	PROPERTY(int, Left);
	GET(int, Left);
	SET(int, Left);
	PROPERTY(int, Top);
	GET(int, Top);
	SET(int, Top);

	PROPERTY(int, X);
	GET(int, X);
	SET(int, X);
	PROPERTY(int, Y);
	GET(int, Y);
	SET(int, Y);

	PROPERTY(int, Right);
	GET(int, Right);
	SET(int, Right);
	PROPERTY(int, Bottom);
	GET(int, Bottom);
	SET(int, Bottom);


	PROPERTY(POINT, Location);
	GET(POINT, Location);
	SET(POINT, Location);
	PROPERTY(SIZE, Size);
	GET(SIZE, Size);
	SET(SIZE, Size);

	PROPERTY(BOOL, Checked);
	GET(BOOL, Checked);
	SET(BOOL, Checked);

	PROPERTY(RECT, Rect);
	GET(RECT, Rect);
	SET(RECT, Rect);

	PROPERTY(BOOL, Enable);
	GET(BOOL, Enable);
	SET(BOOL, Enable);

	PROPERTY(BOOL, Visable);
	GET(BOOL, Visable);
	SET(BOOL, Visable);

	PROPERTY(COLORREF, ForeColor);
	virtual GET(COLORREF, ForeColor);
	virtual SET(COLORREF, ForeColor);

	PROPERTY(COLORREF, BackColor);
	virtual GET(COLORREF, BackColor);
	virtual SET(COLORREF, BackColor);

	PROPERTY(std::vector<Control*>, Childens);
	GET(std::vector<Control*>, Childens);

	PROPERTY(std::vector<HWND>, ChildHandles);
	GET(std::vector<HWND>, ChildHandles);

	PROPERTY(HFONT, Font);
	GET(HFONT, Font);
	SET(HFONT, Font);

	PROPERTY(HBITMAP, Bitmap);
	GET(HBITMAP, Bitmap);
	SET(HBITMAP, Bitmap);

	READONLY_PROPERTY(HWND, TopLevelControl);
	GET(HWND, TopLevelHandle);

	READONLY_PROPERTY(class Window*, TopLevelWindow);
	GET(class Window*, TopLevelWindow);
	void SetRect(POINT location, SIZE size);
	void SetRect(int x, int y, int width, int height);
	void CaptureMessage();
	BOOL RemoveChild(Control* child);
	void SetFont(std::wstring fontName, int fontSize);
	virtual LRESULT ProcessEvent(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

};

class Button : public Control
{
public:
	Button(Control* parent);
};
class CheckBox : public Control
{
public:
	CheckBox(Control* parent);
};
class ComboBox : public Control
{
public:
	ComboBox(Control* parent);

	PROPERTY(int, SelectedIndex);
	GET(int, SelectedIndex);
	SET(int, SelectedIndex);
	READONLY_PROPERTY(std::wstring, SelectedItem);
	GET(std::wstring, SelectedItem);
	std::wstring operator[](int index);
	void AddItem(std::wstring index);
	READONLY_PROPERTY(int, Count);
	GET(int, Count);
	void Clear();
	void RemoveAt(int index);
};
class GroupBox : public Control
{
public:
	GroupBox(Control* parent);
};
class Label : public Control
{
public:
	Label(Control* parent);
};
class ListBox : public Control
{
public:
	ListBox(Control* parent);
	void AddItem(std::wstring item);
	void RemoveAt(int index);
	std::wstring operator[](int index);
	void Clear();
	READONLY_PROPERTY(int, Count);
	GET(int, Count);
};
class ListView : public Control
{
public:
	ListView(Control* parent);

	READONLY_PROPERTY(int, ColumnCount);
	GET(int, ColumnCount);
	PROPERTY(int, RowCount);
	GET(int, RowCount);
	SET(int, RowCount);
	void AddColumn(LVCOLUMN col);
	void AddColumn(std::wstring name, int width, DWORD align = LVCFMT_LEFT, DWORD flag = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM);
	void RemoveColumn(int colIndex);
	int AddRow();
	void SetItemText(int col, int row, std::wstring text);
	void Clear();
	PROPERTY(int, SelectedRowIndex);
	GET(int, SelectedRowIndex);
	SET(int, SelectedRowIndex);
	LVITEM GetItem(int colIndex, int rowIndex);

	std::wstring GetItemText(int colIndex, int rowIndex);
};
class ProgressBar : public Control
{
public:
	ProgressBar(Control* parent);
	PROPERTY(int, MaxValue);
	GET(int, MaxValue);
	SET(int, MaxValue);
	PROPERTY(int, MinValue);
	GET(int, MinValue);
	SET(int, MinValue);
	PROPERTY(int, Value);
	GET(int, Value);
	SET(int, Value);

};
class RadioBox : public Control
{
public:
	RadioBox(Control* parent);
};
class RichTextBox : public Control
{
public:
	RichTextBox(Control* parent);
	void AppendText(std::wstring str);
	void AppendLine(std::wstring str);
	void ScrollToCaret();
};
enum class ScrollBarDirection
{
	Horizontal,
	Vertical
};
class ScrollBar : public Control
{
public:
	ScrollBar(Control* parent, ScrollBarDirection dir = ScrollBarDirection::Horizontal);
};
class TabPage : public Control
{
public:
	TabPage(Control* parent);
};
class TabControl : public Control
{
	List<TabPage*> pages = List<TabPage*>();
public:
	TabControl(Control* parent);
	TabPage* AddPage(std::wstring item);
	void RemoveAt(int index);
	std::wstring operator[](int index);
	void Clear();
	LRESULT ProcessEvent(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	void UpdatePages();
	READONLY_PROPERTY(int, Count);
	GET(int, Count);
	READONLY_PROPERTY(int, SelectedIndex);
	GET(int, SelectedIndex);
	READONLY_PROPERTY(TabPage**, Pages);
	GET(TabPage**, Pages);
};
class TextBox : public Control
{
#define ID_EDITCHILD 100
public:
	TextBox(Control* parent);
	PROPERTY(const wchar_t, PasswordChar);
	GET(const wchar_t, PasswordChar);
	SET(const wchar_t, PasswordChar);
	PROPERTY(BOOL, PasswordBox);
	GET(BOOL, PasswordBox);
	SET(BOOL, PasswordBox);
};
class Tooltip : public Control
{
	Control* bindControl = NULL;
public:
	Tooltip(Control* parent);
	void SetMessage(std::wstring text);
};
class Trackbar : public Control
{
public:
	Trackbar(Control* parent);
	PROPERTY(int, MaxValue);
	GET(int, MaxValue);
	SET(int, MaxValue);
	PROPERTY(int, MinValue);
	GET(int, MinValue);
	SET(int, MinValue);
	PROPERTY(int, Value);
	GET(int, Value);
	SET(int, Value);
	PROPERTY(int, SelStart);
	GET(int, SelStart);
	SET(int, SelStart);
	PROPERTY(int, SelEnd);
	GET(int, SelEnd);
	SET(int, SelEnd);
};
class TreeView : public Control
{
public:
	TreeView(Control* parent);
	void Clear();
	HTREEITEM AddItem(const std::wstring& text, HTREEITEM parent = TVI_ROOT);
	BOOL RemoveItem(HTREEITEM it);

	READONLY_PROPERTY(HTREEITEM, Selected);
	GET(HTREEITEM, Selected);
	READONLY_PROPERTY(std::wstring, SelectedString);
	GET(std::wstring, SelectedString);
	READONLY_PROPERTY(int, Count);
	GET(int, Count);
	std::wstring GetItemString(HTREEITEM it);

};
class DatePicker : public Control
{
public:
	DatePicker(Control* parent);
};
class TimePicker : public Control
{
public:
	TimePicker(Control* parent);
};
class Window : public Control
{
	static BOOL ClassInited;
	static void InitWindowClass();
	COLORREF m_BackColor = RGB(255, 255, 255);
public:
	Window();
	void Show();
	void ShowDialog(HWND parent = NULL);
	static void Run();
	static void DoEvent();
	virtual GET(COLORREF, BackColor);
	virtual SET(COLORREF, BackColor);
	PROPERTY(HICON, Icon);
	GET(HICON, Icon);
	SET(HICON, Icon);
	PROPERTY(BOOL, CanResize);
	GET(BOOL, CanResize);
	SET(BOOL, CanResize);
	PROPERTY(BOOL, MaxBox);
	GET(BOOL, MaxBox);
	SET(BOOL, MaxBox);
	PROPERTY(BOOL, MinBox);
	GET(BOOL, MinBox);
	SET(BOOL, MinBox);
	static int g_WindowCount;
	static List<HWND> Windows;
	void Close();
};
#define GUI_EVENT_LINK(x) [&](Control* sender, HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {x(sender, hWnd, message, wParam, lParam); }